/**
 * Copyright (C) 2008 Gauvain Pocentek <gpocentek@linutop.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif

#include <X11/XKBlib.h>
#include <gdk/gdkx.h>
#include <gtk/gtk.h>
#include <libxfce4util/libxfce4util.h>
#include <libxfcegui4/libxfcegui4.h>
#include <libxfce4panel/xfce-panel-plugin.h>
#include <libxfce4panel/xfce-hvbox.h>

#define CAPSOFF PIXMAPSDIR "capslock_off.xpm"
#define CAPSON PIXMAPSDIR "capslock_on.xpm"
#define NUMOFF PIXMAPSDIR "numlock_off.xpm"
#define NUMON PIXMAPSDIR "numlock_on.xpm"

/*
 * NOTE: tested on debian lenny, probably broken on other versions/distros
 */

enum
{
    CAPS  = 0,
    NUM
};

typedef struct _LedsPlugin
{
    XfcePanelPlugin *plugin;
    GtkWidget *ebox;

    GtkWidget *box;
    GtkWidget *caps_img;
    GtkWidget *num_img;

    GdkPixbuf *caps_pxb_off;
    GdkPixbuf *caps_pxb_on;
    GdkPixbuf *num_pxb_off;
    GdkPixbuf *num_pxb_on;

    int xkbev;
} LedsPlugin;

Display *dpy;

static void
leds_construct (XfcePanelPlugin *plugin);

/* register the plugin */
XFCE_PANEL_PLUGIN_REGISTER_EXTERNAL (leds_construct);

static void
update_icons (LedsPlugin *leds,
              int caps_state,
              int num_state)
{
    if (caps_state)
    {
        gtk_image_set_from_pixbuf (GTK_IMAGE (leds->caps_img),
            leds->caps_pxb_on);
        gtk_widget_set_tooltip_text (leds->caps_img,
            _("Caps Lock is On"));
    } else {
        gtk_image_set_from_pixbuf (GTK_IMAGE (leds->caps_img),
            leds->caps_pxb_off);
        gtk_widget_set_tooltip_text (leds->caps_img,
            _("Caps Lock is Off"));
    }

    if (num_state)
    {
        gtk_image_set_from_pixbuf (GTK_IMAGE (leds->num_img),
            leds->num_pxb_on);
        gtk_widget_set_tooltip_text (leds->num_img,
            _("Num Lock is On"));
    } else {
        gtk_image_set_from_pixbuf (GTK_IMAGE (leds->num_img),
            leds->num_pxb_off);
        gtk_widget_set_tooltip_text (leds->num_img,
            _("Num Lock is Off"));
    }
}

static GdkFilterReturn
event_filter(GdkXEvent *gdkxevent, GdkEvent *event, gpointer data)
{
    LedsPlugin *leds = (LedsPlugin *) data;
    XkbEvent ev;
    int num_state, caps_state;

    memcpy(&ev.core, gdkxevent, sizeof(ev.core));
    if (ev.core.type == leds->xkbev + XkbEventCode)
        if (ev.any.xkb_type == XkbIndicatorStateNotify)
        {
            caps_state = ev.indicators.state & (1 << CAPS) ? 1 : 0;
            num_state = ev.indicators.state & (1 << NUM) ? 1 : 0;
            update_icons (leds, caps_state, num_state);
        }

    return GDK_FILTER_CONTINUE;
}

/* Panel plugin callbacks */
void
leds_save (XfcePanelPlugin *plugin,
             LedsPlugin    *leds)
{
    /* nothing */
}

static LedsPlugin *
leds_new (XfcePanelPlugin *plugin)
{
    LedsPlugin *leds;
    GtkOrientation orientation;
    leds = panel_slice_new0 (LedsPlugin);
    leds->plugin = plugin;

    /* kkb extension init */
    int maj, min, code;
    int xkberr;
    GdkDrawable *drawable;

    maj = XkbMajorVersion;
    min = XkbMinorVersion;
    drawable = gdk_get_default_root_window();
    dpy = gdk_x11_drawable_get_xdisplay(drawable);

    XkbQueryExtension(dpy, &code, &leds->xkbev, &xkberr, &maj, &min);
    XkbSelectEvents(dpy, XkbUseCoreKbd, XkbIndicatorStateNotifyMask, XkbIndicatorStateNotifyMask);

    /* plugin */
    leds->ebox = gtk_event_box_new ();

    orientation = xfce_panel_plugin_get_orientation (plugin);
    leds->box = xfce_hvbox_new (orientation, TRUE, 1);
    gtk_container_add (GTK_CONTAINER (leds->ebox), leds->box);

    leds->caps_img = gtk_image_new ();
    leds->num_img = gtk_image_new ();
    gtk_box_pack_start (GTK_BOX (leds->box), leds->caps_img, FALSE, FALSE, 0);
    gtk_box_pack_start (GTK_BOX (leds->box), leds->num_img, FALSE, FALSE, 0);

    gtk_widget_show_all (leds->ebox);

    /* keeps the icons in memory */
    leds->caps_pxb_off = gdk_pixbuf_new_from_file (CAPSOFF, NULL);
    leds->caps_pxb_on = gdk_pixbuf_new_from_file (CAPSON, NULL);
    leds->num_pxb_off = gdk_pixbuf_new_from_file (NUMOFF, NULL);
    leds->num_pxb_on = gdk_pixbuf_new_from_file (NUMON, NULL);

    /* watch keyboard events */
    gdk_window_add_filter(NULL, (GdkFilterFunc)event_filter, (gpointer) leds);

    /* initial icon state */
    int ledstate, num, caps;
    XkbGetIndicatorState(dpy, XkbUseCoreKbd, &ledstate);
    num = ledstate & 1<<NUM;
    caps = ledstate & 1<<CAPS;
    update_icons (leds, caps, num);

    return leds;
}

static void
leds_free (XfcePanelPlugin *plugin,
             LedsPlugin    *leds)
{
    gtk_widget_destroy (leds->ebox);
    panel_slice_free (LedsPlugin, leds);
}

static void
leds_orientation_changed (XfcePanelPlugin *plugin,
                            GtkOrientation   orientation,
                            LedsPlugin    *leds)
{
    xfce_hvbox_set_orientation (XFCE_HVBOX (leds->box), orientation);
}

static gboolean
leds_size_changed (XfcePanelPlugin *plugin,
                     gint             size,
                     LedsPlugin    *leds)
{
    /* nothing to do */
}

static void
leds_construct (XfcePanelPlugin *plugin)
{
    xfce_textdomain(GETTEXT_PACKAGE, PACKAGE_LOCALE_DIR, "UTF-8");

    LedsPlugin *leds;
    leds = leds_new (plugin);
    gtk_container_add (GTK_CONTAINER (plugin), leds->ebox);
    xfce_panel_plugin_add_action_widget (plugin, leds->ebox);

    g_signal_connect (G_OBJECT (plugin), "free-data",
                      G_CALLBACK (leds_free), leds);
    g_signal_connect (G_OBJECT (plugin), "save",
                      G_CALLBACK (leds_save), leds);
    g_signal_connect (G_OBJECT (plugin), "size-changed",
                      G_CALLBACK (leds_size_changed), leds);
    g_signal_connect (G_OBJECT (plugin), "orientation-changed",
                      G_CALLBACK (leds_orientation_changed), leds);

/*
    xfce_panel_plugin_menu_show_about (plugin);
    g_signal_connect (G_OBJECT (plugin), "about",
                      G_CALLBACK (leds_about), NULL);
*/
}

